package LuxSci_API2;

use LWP::UserAgent;
use MIME::Base64;
use Digest::SHA;
use JSON::XS;
use strict;

=head1 name

LuxSci_API2

=head1 description

Client-side interface to calling LuxSci v2 API functions.

=head1 methods

=over 4

=item $api = new LuxSci_API2(%p);

Create an API2 object and Authenticate.

%p:
   token:  Required. API Token
   secret: Required. API Secret
	host:   Optinal. API Hostname (default is rest.luxsci.com)
   debug:  0|1, If true, emit debugging telemetry
   user:   Optional username
	pass:   Optional password

Returns the $api object.  Use getLastResponse to get the results of the auth call.

=cut

sub new {
	my ($class,%data) = @_;

	my $this = \%data;
	
	$$this{path}   = "/perl/api/v2/auth";
	$$this{host} ||= "rest.luxsci.com";
	
	$this    =  bless ( $this, $class );

	my $req  = HTTP::Request->new("POST", "https://$$this{host}$$this{path}");
	my $date = time();
	my $json = undef;

	if ($$this{user}) {
  		$json = {	
			token       => $$this{token},
			date        => $date,
			user        => $$this{user},
			pass        => $$this{pass},
			signature   => Digest::SHA::hmac_sha256_hex( "$$this{token}\n$date\n$$this{user}\n$$this{pass}\n", $$this{secret} ),
	 		};
		}
	else {
  		$json = {	
			token       => $$this{token},
			date        => $date,
			signature   => Digest::SHA::hmac_sha256_hex( "$$this{token}\n$date\n", $$this{secret} ),
	 		};
		}

	eval {
		 $req->header("content-type" => "application/json");
		 $req->content( JSON::XS::encode_json($json) );
		 $this->_callAPI( $req );
		 };
	if ($@) {
		 $this->setLastResponse( 400,  { success => 0, error => "Failure encoding JSON request: $@" } );
		 return $this; 
		 }

	return $this;
	}


##
## ###################################################################################################
##

## Useful in account-scope APIs to set the username for the API path
sub setUser {
	my ($api, $username) = @_;
	$$api{user} = $username;
	}

sub setDebugFlag {
	my ($api, $debug_flag) = @_;
	$$api{debug} = int($debug_flag);
	}

sub clearLastResponse {
	my ($this) = @_;
	$this->setLastResponse( 0, {} );
	}

sub setLastResponse {
	my ($this,$code,$response) = @_;
	return $$this{last_response} = [ $code, $response ];
	}

sub getLastResponse {
	my $this = shift;	
	return $$this{last_response} || [ 0, {} ];
	}


##
## ###################################################################################################
##

=item $api->RevokeSession(%parms)

Revoke an active API session.

=cut

sub RevokeSession {
	my ($this) = @_;
	$this->call( method => "DELETE", path => "/perl/api/v2/auth" );
	}


sub SendEmail {
	my ($this, %p) = @_;
	
	$this->call( 
		method             => "POST", 
		path               => "/perl/api/v2/user/$$this{user}/email/send",
		data               => {
			messages           => $p{messages},
			smtp_server        => $p{smtp_server},
			smtp_server_method => int($p{smtp_server_method}) || 2,
			},
		files              => $p{files}
		);
	}


#######################################################

=item $api->call(%p);

Execute an api function.  The $api object must be already authenticated.

%p:
	method: (Optional; Default is GET). HTTP Method to use.  E.g. GET, PUT, POST, ...
	path:   path of the request. E.g. "/perl/api/v2/user/username/settings/timezones"
	data:   Hash reference of data to send. Required to at least be empty has href: {}
	
Data will be turned into a JSON for PUT/POST calls, and a query string
for all others.

=cut

sub call {
	my ($this,%p) = @_;
	my $auth   = $$this{auth};
	my $method = uc($p{method} || "GET");
	my $path   = $p{path} || "/perl/api/v2";
	my $files  = $p{files};
	my $qs     = "";
	my $json   = undef;
	my $hmac   = undef;

	if ($method =~ /PUT|POST/) {
		eval {
			$json = JSON::XS::encode_json($p{data});
			};
		if ($@) {
			$this->setLastResponse(400, { success => 0, error => "Failure encoding JSON request: $@" });
			return;
			}

		# Must be sure to be trimmed before applying hmac.
		$json =~ s/^\s+|\s+$//g;
		$hmac = Digest::SHA::sha256_hex( $json );
		}

	else {
		my @list = ();
		for my $key (keys(%{$p{data}})) {
			push @list, URL_Encode($key) . "=" . URL_Encode($p{data}{$key});
			}
		$qs = join("&", @list);
		}

	 my $req = HTTP::Request->new("GET", "https://$$this{host}" . $p{path} . ($qs ? "?$qs" : ""));
	 $req->method($method);

    # Send the POST/PUT body
    if ($method eq "POST" && ref($files)) {
	 	 my $boundary = "-----------------LUXSCI-" . rand(999999) . "---";
		 my $body = "";
	 
		 $req->header("content-type" => "multipart/form-data; boundary=$boundary");		 
		 
		 if ($json) {
			 $body .= "--${boundary}\r\nContent-Disposition: form-data; name=\"json\"; filename=\"json.js\"\r\nContent-Type: application/json\r\n\r\n";
			 $body .=  $json . "\r\n";
			 }

		 for my $file (@$files) {
			 my $filename = $$file{filename};
			 $body .= "--${boundary}\r\nContent-Disposition: form-data; name=\"files\"; filename=\"$filename\"\r\n";

			 my $ct = CGIFiles::getContentType( $filename );
			 $body .= "Content-Type: $ct\r\n";

			 $body .= "\r\n";
			 $body .= ref($$file{data}) ? ${ $$file{data} } : $$file{data};
			 $body .= "\r\n";
			 }
			 
		 $body .= "--${boundary}--\r\n";
		 $req->content( $body );
       }

	 # Send the POST/PUT body
	 elsif ($json) {
		 $req->header("content-type" => "application/json");
		 $req->content( $json );
		 }

	 # Request signature	
    my $to_sign = "$auth\n" . 
	   uc($method) . "\n" .
		$path . "\n" .
		$qs   . "\n" .
		$hmac . "\n";

	 my $sig = Digest::SHA::hmac_sha256_hex( $to_sign, $$this{secret} );

    $req->header("Cookie" => "signature=$auth:$sig");

	 $this->_callAPI( $req );
	 }



#######################################################

=item $api->_callAPI($request);

Back-end function to actually perform the api request.

=cut

sub _callAPI {
	my ($this, $req) = @_;
	my $debug = $$this{debug};

	my $ua = LWP::UserAgent->new(timeout => 120, agent => "LuxSci APIv2 Perl Client");

	if ($debug) {
	   $ua->add_handler("request_send", 
			sub { print STDERR "\n\nREQUEST:----\n" . shift->as_string; return });

   	$ua->add_handler("response_done", 
			sub { my $r = shift->as_string; print STDERR "\n\nRESPONSE:----\n$r"; return });
		}

	 my $response = $ua->request( $req );
	 my $code     = $response->code;
	 my $data     = undef;
	 
	 eval {
	 	 $data = parseResult( $response->as_string );
		 };

	 if ($@) {
		 if ($debug) {
		    print "JSON Parse Exception: $@\n";
			 }		 	
		 $this->setLastResponse( 500, { success => 0, error => "Non-JSON or unparsable data returned." } );
		 }
	 else {
		 $this->setLastResponse( $code, $data );
		 }	 	

	 # Save [new] authentication token
	 if ($$data{success}) {
		 $$this{auth} = $$data{auth};
		 }
	 }



###########################################################


# Parse HTTP Response and JSON data returned.
sub parseResult {
	my $string   = shift;
	my @response = split(/\n/, $string);

	my $data = "";
	for (my $i=0; $i <= $#response; $i++) {
		if ($response[$i] =~ /^\s*$/) {
			$data = join("\n", @response[($i+1)..$#response]);
			last;
			}
		}

	return JSON::XS::decode_json( $data );
	}

# URL-encode text
sub URL_Encode {
   my ($a) = shift;
   $a =~ s/([+?&=#<> %"']|[^\x20-\x7E])/"%".to_hex(ord($1))/eg;
   return $a;
   }

# Convert number to 2-digit HEX
sub to_hex {
   my $n = shift;
   my @h = qw( 0 1 2 3 4 5 6 7 8 9 a b c d e f);
   return $h[ int($n/16) ] . $h[ int($n % 16) ];
   }

1;
